from typing import List
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from pydantic import NoneIsAllowedError
from supervisely.fastapi_helpers.singleton import Singleton


class WebsocketManager(metaclass=Singleton):
    def __init__(self, path="/sly-app-ws"):
        self.app = None
        self.path = path
        self.active_connections: List[WebSocket] = []
    
    def set_app(self, app: FastAPI):
        if self.app is not None:
            return
        self.app = app
        self.app.add_api_websocket_route(path=self.path, endpoint=self.endpoint)

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def broadcast(self, d: dict): 
        if self.app is None:
            raise ValueError("WebSocket is not initialized, use Websocket middleware for that")
        for connection in self.active_connections:
            await connection.send_json(d)

    async def endpoint(self, websocket: WebSocket):
        await self.connect(websocket)
        try:
            while True:
                data = await websocket.receive_text()
        except WebSocketDisconnect:
            self.disconnect(websocket)



# coding: utf-8

import cv2
import numpy as np

from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list
from supervisely.geometry import validation


# @TODO: validation
class Rectangle(Geometry):
    '''
    This is a class for creating and using Rectangle objects for Labels
    '''
    @staticmethod
    def geometry_name():
        return 'rectangle'

    def __init__(self, top, left, bottom, right,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """
        Float-type coordinates will be deprecated soon.
        Args:
            top: minimal vertical value
            left: minimal horizontal value
            bottom: maximal vertical value
            right: maximal horizontal value
        """

        if top > bottom:
            raise ValueError('Rectangle "top" argument must have less or equal value then "bottom"!')

        if left > right:
            raise ValueError('Rectangle "left" argument must have less or equal value then "right"!')

        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at,
                         created_at=created_at)

        self._points = [PointLocation(row=top, col=left), PointLocation(row=bottom, col=right)]

    """
    Implementation of all methods from Geometry
    """

    def to_json(self):
        '''
        The function to_json convert Rectangle class object to json format
        :return: Rectangle in json format
        '''
        packed_obj = {
            POINTS: {
                EXTERIOR: points_to_row_col_list(self._points, flip_row_col_order=True),
                INTERIOR: []
            }
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Rectangle from json format to Rectangle class object. If json format is not correct it generate exception error.
        :param data: input Rectangle in json format
        :return: Rectangle class object
        '''
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)

        exterior = data[POINTS][EXTERIOR]
        if len(exterior) != 2:
            raise ValueError('"exterior" field must contain exactly two points to create Rectangle object.')
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])
        return cls(top=top, left=left, bottom=bottom, right=right,
                   sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def crop(self, other):
        '''
        Crop the current Rectangle with a given rectangle
        :param other: Rectangle class object
        :return: list with Rectangle class object, if Rectangle class object not intersect with given rectangle, empty list will be returned
        '''
        top = max(self.top, other.top)
        left = max(self.left, other.left)
        bottom = min(self.bottom, other.bottom)
        right = min(self.right, other.right)
        is_valid = (bottom >= top) and (left <= right)
        return [Rectangle(top=top, left=left, bottom=bottom, right=right)] if is_valid else []

    def _transform(self, transform_fn):
        transformed_corners = [transform_fn(p) for p in self.corners]
        rows, cols = zip(*points_to_row_col_list(transformed_corners))
        return Rectangle(top=round(min(rows)), left=round(min(cols)), bottom=round(max(rows)), right=round(max(cols)))

    @property
    def corners(self):
        '''
        :return: list of coners(PointLocation class objects) of Rectangle
        '''
        return [PointLocation(row=self.top, col=self.left), PointLocation(row=self.top, col=self.right),
                PointLocation(row=self.bottom, col=self.right), PointLocation(row=self.bottom, col=self.left)]

    def rotate(self, rotator):
        '''
        The function rotate Rectangle with a given rotator(ImageRotator class object contain size of image and angle to rotate)
        :param rotator: ImageRotator class object
        :return: Rectangle class object
        '''
        return self._transform(lambda p: rotator.transform_point(p))

    def resize(self, in_size, out_size):
        '''
        Resize the current Rectangle to match a certain size
        :param in_size: input image size
        :param out_size: output image size
        :return: Rectangle class object
        '''
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor):
        '''
        The function scale change scale of the current Rectangle object with a given factor
        :param factor: float scale parameter
        :return: Rectangle class object
        '''
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate shifts the rectangle by a certain number of pixels and return the copy of the current Rectangle object
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: Rectangle class object
        '''
        return self._transform(lambda p: p.translate(drow, dcol))

    def fliplr(self, img_size):
        '''
        The function fliplr the current Rectangle object geometry in horizontal
        :param img_size: size of the image
        :return: Rectangle class object
        '''
        img_width = img_size[1]
        return Rectangle(top=self.top, left=(img_width - self.right), bottom=self.bottom, right=(img_width - self.left))

    def flipud(self, img_size):
        '''
        The function flipud the current Rectangle object geometry in vertical
        :param img_size: size of the image
        :return: Rectangle class object
        '''
        img_height = img_size[0]
        return Rectangle(top=(img_height - self.bottom), left=self.left, bottom=(img_height - self.top),
                         right=self.right)

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        self._draw_contour_impl(bitmap, color, thickness=cv2.FILLED, config=config)  # due to cv2

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        cv2.rectangle(bitmap, pt1=(self.left, self.top), pt2=(self.right, self.bottom), color=color,
                      thickness=thickness)

    def to_bbox(self):
        '''
        :return: copy of current Rectangle class object
        '''
        return self.clone()

    @property
    def area(self):
        '''
        :return: area of current Rectangle object
        '''
        return float(self.width * self.height)

    @classmethod
    def from_array(cls, arr):
        '''
        The function from_array create Rectangle object with given array shape
        :param arr: numpy array
        :return: Rectangle class object
        '''
        return cls(top=0, left=0, bottom=arr.shape[0] - 1, right=arr.shape[1] - 1)

    # TODO re-evaluate whether we need this, looks trivial.
    @classmethod
    def from_size(cls, size: tuple):
        '''
        The function from_size create Rectangle object with given size shape
        :param size: tuple of integers
        :return: Rectangle class object
        '''
        return cls(0, 0, size[0] - 1, size[1] - 1)

    @classmethod
    def from_geometries_list(cls, geometries):
        '''
        The function from_geometries_list create Rectangle object from given geometry object
        :param geometries: list of geometry type objects(Point, Polygon, PolyLine, Bitmap etc.)
        :return: Rectangle class object
        '''
        bboxes = [g.to_bbox() for g in geometries]
        top = min(bbox.top for bbox in bboxes)
        left = min(bbox.left for bbox in bboxes)
        bottom = max(bbox.bottom for bbox in bboxes)
        right = max(bbox.right for bbox in bboxes)
        return cls(top=top, left=left, bottom=bottom, right=right)

    @property
    def left(self):
        return self._points[0].col

    @property
    def right(self):
        return self._points[1].col

    @property
    def top(self):
        return self._points[0].row

    @property
    def bottom(self):
        return self._points[1].row

    @property
    def center(self):
        '''
        :return: center of rectangle(PointLocation class obgect)
        '''
        return PointLocation(row=(self.top + self.bottom) // 2, col=(self.left + self.right) // 2)

    @property
    def width(self):
        '''
        :return: width of rectangle(int)
        '''
        return self.right - self.left + 1

    @property
    def height(self):
        '''
        :return: height of rectangle(int)
        '''
        return self.bottom - self.top + 1

    def contains(self, rect):
        '''
        The function contains checks if Rectangle class object contains a given rectangle
        :param rect: Rectangle class object
        :return: bool
        '''
        return (self.left <= rect.left and
                self.right >= rect.right and
                self.top <= rect.top and
                self.bottom >= rect.bottom)

    def contains_point_location(self, pt: PointLocation):
        '''
        The function contains_point_location checks if Rectangle class object contains a given point
        :param pt: PointLocation class object
        :return: bool
        '''
        return (self.left <= pt.col <= self.right) and (self.top <= pt.row <= self.bottom)

    def to_size(self):
        '''
        :return: height and width of rectangle(int)
        '''
        return self.height, self.width

    def get_cropped_numpy_slice(self, data: np.ndarray) -> np.ndarray:
        '''
        The function get_cropped_numpy_slice checks slice of given numpy array with Rectangle parameters
        :param data: numpy array
        :return: numpy array
        '''
        return data[self.top:(self.bottom+1), self.left:(self.right+1), ...]

    def intersects_with(self, rect):
        if self.left > rect.right or self.right < rect.left:
            return False
        if self.top > rect.bottom or self.bottom < rect.top:
            return False
        return True

    @classmethod
    def allowed_transforms(cls):
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon
        return [AnyGeometry, Bitmap, Polygon]



# coding: utf-8
import urllib.parse

from supervisely.api.labeling_job_api import LabelingJobApi
Status = LabelingJobApi.Status
import supervisely.labeling_jobs.constants as constants
from supervisely.api.module_api import ApiField


def total_desc():
    return "TOTAL", "the total number of jobs in current team"


def is_completed_desc():
    return "COMPLETED", "the number of completed jobs"


def is_completed(job_info):
    if job_info.status == str(Status.COMPLETED):
        return True


def is_stopped_desc():
    return "STOPPED", "the number of stopped jobs"


def is_stopped(job_info):
    if job_info.status == str(Status.STOPPED):
        return True


def is_not_started_desc():
    return "PENDING", "the number of jobs labeler haven't even opened yet (created but not started)"


def is_not_started(job_info):
    if job_info.status == str(Status.PENDING):
        return True


def total_items_count_desc():
    return "TOTAL", "the total number of items in all labeling jobs"


def total_items_count(job_info):
    return job_info.images_count


def labeled_items_count_desc():
    return "LABELED", "the total number of labeled items (labelers marked as \"finished\") in all labeling jobs"


# cnt images, that labeler marked as done
def labeled_items_count(job_info):
    if is_on_labeling(job_info):
        return job_info.finished_images_count
    else:
        return total_items_count(job_info)


def reviewed_items_count_desc():
    return "REVIEWED", "the total number of reviewed items (reviewers marked as \"accepted\" or \"rejected\") in all labeling jobs"


# cnt images, that reviewer accepted or rejected
def reviewed_items_count(job_info):
    return job_info.accepted_images_count + job_info.rejected_images_count


def accepted_items_count_desc():
    return "ACCEPTED", "the total number of accepted items (reviewers marked as \"accepted\") in all labeling jobs"


def accepted_items_count(job_info):
    return job_info.accepted_images_count


def rejected_items_count_desc():
    return "REJECTED", "the total number of rejected items (reviewers marked as \"rejected\") in all labeling jobs"


def rejected_items_count(job_info):
    return job_info.rejected_images_count


def is_on_labeling_desc():
    return 'LABELING IN PROGRESS', "the number of jobs with status IN_PROGRESS"


# labeling is in progress
def is_on_labeling(job_info):
    if job_info.status == str(Status.IN_PROGRESS):
        return True
    return False


def is_labeling_started_desc():
    return 'LABELING STARTED', "the number of jobs that are started by labeler and with at least one labeled item (marked \"done\" by labeler)"


# cnt jobs that are started by labeler and with at least one image that marked "done" by labeler
def is_labeling_started(job_info):
    if is_on_labeling(job_info) and labeled_items_count(job_info) != 0:
        return True
    return False


def is_on_review_desc():
    return 'ON REVIEW', "the number of jobs with status ON_REVIEW"


# cnt jobs with "on review" status
def is_on_review(job_info):
    if job_info.status == str(Status.ON_REVIEW):
        return True
    return False


def is_review_started_desc():
    return 'REVIEW STARTED', "the number of jobs that are started by reviewer - with at least one reviewed item (marked \"accepted\" or \"rejected\")"


# cnt jobs with at least one reviewed (accepted or rejected) item
def is_review_started(job_info):
    if is_on_review(job_info) and reviewed_items_count(job_info) != 0:
        return True
    return False


def is_zero_labeling_desc():
    return 'ZERO LABELED', "the number of jobs with status \"IN PROGRESS\" with zero labeled items"


def is_zero_reviewed_desc():
    return 'ZERO REVIEWED', "the number of jobs with status \"ON REVIEW\" with zero reviewed items"


def get_job_url(server_address, job):
    result = urllib.parse.urljoin(server_address, 'app/images/{}/{}/{}/{}?jobId={}'.format(job.team_id,
                                                                                          job.workspace_id,
                                                                                          job.project_id,
                                                                                          job.dataset_id,
                                                                                          job.id))

    return result


def jobs_stats(server_address, jobs, stats):
    import pandas as pd
    col_job_id = []
    col_job_name = []  # link here
    col_job_status = []
    col_items_total = []
    col_items_labeled = []
    col_items_reviewed = []
    col_items_accepted = []
    col_items_rejected = []
    col_created_at = []

    for job, stat in zip(jobs, stats):
        col_job_id.append(job.id)
        col_job_name.append('<a href="{0}" rel="noopener noreferrer" target="_blank">{1}</a>'
                            .format(get_job_url(server_address, job), job.name))
        col_job_status.append(job.status)
        col_created_at.append(job.created_at)
        col_items_total.append(total_items_count(job))
        col_items_labeled.append(labeled_items_count(job))
        col_items_reviewed.append(reviewed_items_count(job))
        col_items_accepted.append(accepted_items_count(job))
        col_items_rejected.append(rejected_items_count(job))

    df = pd.DataFrame(list(zip(col_job_id,
                               col_job_name,
                               col_job_status,
                               col_items_total,
                               col_items_labeled,
                               col_items_reviewed,
                               col_items_accepted,
                               col_items_rejected,
                               col_created_at)),
                      columns=['ID', 'NAME', 'STATUS', 'TOTAL', 'LABELED', 'REVIEWED', 'ACCEPTED', 'REJECTED', 'CREATED_AT'])

    df['CREATED_AT'] = pd.to_datetime(df['CREATED_AT']).dt.strftime('%d/%m/%Y %H:%M')
    return df


def jobs_summary(jobs):
    import pandas as pd
    count_total = len(jobs)

    count_completed = 0
    count_stopped = 0

    count_labeling_not_started = 0  # pending

    count_on_labeling = 0
    count_labeling_started = 0
    count_labeling_zero_done = 0

    count_on_review = 0
    count_review_started = 0
    count_review_zero_done = 0

    for job in jobs:
        if is_completed(job):
            count_completed += 1
        elif is_stopped(job):
            count_stopped += 1
        elif is_not_started(job):
            count_labeling_not_started += 1
        elif is_on_labeling(job):
            count_on_labeling += 1
            if is_labeling_started(job):
                count_labeling_started += 1
        elif is_on_review(job):
            count_on_review += 1
            if is_review_started(job):
                count_review_started += 1
        else:
            raise RuntimeError("Unhandled job status: {}".format(str(job)))

    count_labeling_zero_done = count_on_labeling - count_labeling_started
    count_review_zero_done = count_on_review - count_review_started

    names = []
    percentages = []
    descriptions = []

    table_rows_f = [total_desc, is_completed_desc, is_stopped_desc,
                    is_not_started_desc,
                    is_on_labeling_desc, is_labeling_started_desc, is_zero_labeling_desc,
                    is_on_review_desc, is_review_started_desc, is_zero_reviewed_desc]

    values = [count_total,
              count_completed,
              count_stopped,

              count_labeling_not_started,

              count_on_labeling,
              count_labeling_started,
              count_labeling_zero_done,

              count_on_review,
              count_review_started,
              count_review_zero_done,
              ]

    for v, func in zip(values, table_rows_f):
        percentages.append("{:.2f} %".format(v * 100 / count_total))
        name, desc = func()
        names.append(name)
        descriptions.append(desc)

    df = pd.DataFrame(list(zip(list(range(len(names))), names, values, percentages, descriptions)),
                      columns=['#', 'JOB STATUS', 'QUANTITY', 'PERCENTAGE', 'DESCRIPTION'])
    return df


def images_summary(jobs):
    import pandas as pd
    count_total_items = 0
    count_labeled_items = 0
    count_reviewed_items = 0
    count_accepted_items = 0
    count_rejected_items = 0

    for job in jobs:
        count_total_items += total_items_count(job)
        count_labeled_items += labeled_items_count(job)
        count_reviewed_items += reviewed_items_count(job)
        count_accepted_items += accepted_items_count(job)
        count_rejected_items += rejected_items_count(job)

    values_items = [count_total_items,
                    count_labeled_items,
                    count_reviewed_items,
                    count_accepted_items,
                    count_rejected_items]

    items_f = [total_items_count_desc,
               labeled_items_count_desc,
               reviewed_items_count_desc,
               accepted_items_count_desc,
               rejected_items_count_desc]

    names_items = []
    percentages_items = []
    descriptions_items = []

    for v, func in zip(values_items, items_f):
        percentages_items.append("{:.2f} %".format(v * 100 / count_total_items))
        name, desc = func()
        names_items.append(name)
        descriptions_items.append(desc)

    df = pd.DataFrame(list(zip(list(range(len(names_items))), names_items, values_items, percentages_items, descriptions_items)),
                      columns=['#', 'ITEM STATUS', 'QUANTITY', 'PERCENTAGE', 'DESCRIPTION'])
    return df


def classes_summary(stats):
    import pandas as pd
    class_id_stats = {}
    for stat in stats:
        for class_stat in stat[constants.CLASSES_STATS]:
            class_id = class_stat[ApiField.ID]
            class_name = class_stat[ApiField.NAME]
            class_shape = class_stat[ApiField.SHAPE]
            image_count = class_stat[ApiField.IMAGES_COUNT]
            object_count = class_stat[ApiField.LABELS_COUNT]
            class_color = class_stat[ApiField.COLOR]

            if class_id not in class_id_stats:
                class_id_stats[class_id] = {ApiField.NAME: class_name, ApiField.SHAPE: class_shape,
                                            ApiField.COLOR: class_color,
                                            ApiField.IMAGES_COUNT: 0, ApiField.LABELS_COUNT: 0}

            class_id_stats[class_id][ApiField.IMAGES_COUNT] += image_count
            class_id_stats[class_id][ApiField.LABELS_COUNT] += object_count

    col_name = []
    col_shape = []
    col_image_count = []
    col_object_count = []
    for class_id, value in class_id_stats.items():
        col_name.append('<b style="display: inline-block; border-radius: 50%; background: {}; width: 8px; height: 8px"></b> {}'
                        .format(value[ApiField.COLOR], value[ApiField.NAME]))
        col_shape.append(value[ApiField.SHAPE])
        col_image_count.append(value[ApiField.IMAGES_COUNT])
        col_object_count.append(value[ApiField.LABELS_COUNT])

    df = pd.DataFrame(list(zip(list(range(len(col_name))), col_name, col_shape, col_image_count, col_object_count)),
                      columns=['#', 'CLASS', 'SHAPE', 'IMAGES COUNT', 'OBJECTS COUNT'])
    return df


def tags_summary(stats):
    import pandas as pd
    tag_id_stats = {}
    for stat in stats:
        for tag_stat in stat["job"][constants.TAGS_STATS]:
            if "parentId" in tag_stat:
                # @TODO: consider parentId (for oneOf cases)
                continue
            tag_id = tag_stat[ApiField.ID]
            tag_name = tag_stat[ApiField.NAME]
            image_count = tag_stat[ApiField.IMAGES]
            object_count = tag_stat[ApiField.FIGURES]
            class_color = tag_stat[ApiField.COLOR]

            if tag_id not in tag_id_stats:
                tag_id_stats[tag_id] = {ApiField.NAME: tag_name, ApiField.COLOR: class_color,
                                        ApiField.IMAGES_COUNT: 0, ApiField.LABELS_COUNT: 0}

            tag_id_stats[tag_id][ApiField.IMAGES_COUNT] += image_count
            tag_id_stats[tag_id][ApiField.LABELS_COUNT] += object_count

    col_name = []
    col_image_count = []
    col_object_count = []
    for tag_id, value in tag_id_stats.items():
        col_name.append('<b style="display: inline-block; border-radius: 50%; background: {}; width: 8px; height: 8px"></b> {}'
                        .format(value[ApiField.COLOR], value[ApiField.NAME]))
        col_image_count.append(value[ApiField.IMAGES_COUNT])
        col_object_count.append(value[ApiField.LABELS_COUNT])

    df = pd.DataFrame(list(zip(list(range(len(col_name))), col_name, col_image_count, col_object_count)),
                      columns=['#', 'TAG', 'IMAGES COUNT', 'OBJECTS COUNT'])
    return df


import typing
from os import PathLike
import jinja2
from fastapi.templating import Jinja2Templates as _fastapi_Jinja2Templates


class Jinja2Templates(_fastapi_Jinja2Templates):
    def _create_env(
        self, directory: typing.Union[str, PathLike]
    ) -> "jinja2.Environment":
        env_fastapi = super()._create_env(directory)
        env_sly = jinja2.Environment(
            loader=env_fastapi.loader, 
            autoescape=True,
            variable_start_string='{{{',
            variable_end_string='}}}',
        )
        env_sly.globals["url_for"] = env_fastapi.globals["url_for"]
        return env_sly


# coding: utf-8
from collections import namedtuple
from supervisely.api.module_api import ApiField, ModuleApiBase, _get_single_item


class UserApi(ModuleApiBase):
    Membership = namedtuple("Membership", ['id', 'name', 'role_id', 'role'])

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.LOGIN,
                ApiField.ROLE,
                ApiField.ROLE_ID,
                ApiField.NAME,
                ApiField.EMAIL,
                ApiField.LOGINS,
                ApiField.DISABLED,
                ApiField.LAST_LOGIN,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'UserInfo'

    def _convert_json_info(self, info: dict, skip_missing=True):
        return super(UserApi, self)._convert_json_info(info, skip_missing=skip_missing)

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: user metainformation with given id
        '''
        return self._get_info_by_id(id, 'users.info')

    def get_info_by_login(self, login):
        '''
        :param login: str
        :return: user metainformation with given login
        '''
        filters = [{"field": ApiField.LOGIN, "operator": "=", "value": login}]
        items = self.get_list(filters)
        return _get_single_item(items)

    def get_member_info_by_login(self, team_id, login):
        filters = [{"field": ApiField.LOGIN, "operator": "=", "value": login}]
        team_members = self.get_list_all_pages('members.list', {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters},
                                               convert_json_info_cb=self._api.user._convert_json_info)
        return _get_single_item(team_members)

    def get_member_info_by_id(self, team_id, user_id):
        filters = [{"field": ApiField.ID, "operator": "=", "value": user_id}]
        team_members = self.get_list_all_pages('members.list', {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters},
                                               convert_json_info_cb=self._api.user._convert_json_info)
        return _get_single_item(team_members)

    def get_list(self, filters=None):
        '''
        :param filters: list
        :return: list of all registered users
        '''
        return self.get_list_all_pages('users.list', {ApiField.FILTER: filters or []})

    def create(self, login, password, is_restricted=False, name="", email=""):
        '''
        Create new user with given login and password
        :param login: str
        :param password: str
        :param is_restricted: bool
        :param name: str
        :param email: str
        :return: new user metainformation
        '''
        response = self._api.post('users.add', {ApiField.LOGIN: login,
                                                ApiField.PASSWORD: password,
                                                ApiField.IS_RESTRICTED: is_restricted,
                                                ApiField.NAME: name,
                                                ApiField.EMAIL: email,
                                                })
        return self.get_info_by_id(response.json()[ApiField.USER_ID])

    def _set_disabled(self, id, disable):
        '''
        Check status of the user with given id
        :param id: int
        :param disable: bool
        '''
        self._api.post('users.disable', {ApiField.ID: id, ApiField.DISABLE: disable})

    def disable(self, id):
        '''
        Check user with given id is disable
        :param id: int
        '''
        self._set_disabled(id, True)

    def enable(self, id):
        '''
        Check user with given id is enable
        :param id: int
        '''
        self._set_disabled(id, False)

    def get_token(self, login):
        raise NotImplementedError()

    def get_teams(self, id):
        '''
        :param id: int
        :return: list of teams in which given user is a member of
        '''
        response = self._api.post('users.info', {ApiField.ID: id})
        teams_json = response.json()[ApiField.TEAMS]
        teams = []
        for team in teams_json:
            member = self.Membership(id=team[ApiField.ID],
                                     name=team[ApiField.NAME],
                                     role_id=team[ApiField.ROLE_ID],
                                     role=team[ApiField.ROLE])
            teams.append(member)
        return teams

    def add_to_team(self, user_id, team_id, role_id):
        '''
        Invite user to team
        :param user_id: int
        :param team_id: int
        :param role_id: int
        '''
        user = self.get_info_by_id(user_id)
        response = self._api.post('members.add', {ApiField.LOGIN: user.login,
                                                  ApiField.TEAM_ID: team_id,
                                                  ApiField.ROLE_ID: role_id})

    def remove_from_team(self, user_id, team_id):
        '''
        Remove user from team
        :param user_id: int
        :param team_id: int
        '''
        response = self._api.post('members.remove', {ApiField.ID: user_id,
                                                     ApiField.TEAM_ID: team_id})

    def update(self, id, password=None, name=None):
        '''
        Update user info
        :param id: int
        :param password: str
        :param name: str
        :return: updated user metainformation
        '''
        data = {}
        if password is not None:
            data[ApiField.PASSWORD] = password
        if name is not None:
            data[ApiField.NAME] = name
        if len(data) == 0:
            return
        data[ApiField.ID] = id

        self._api.post('users.editInfo', data)
        return self.get_info_by_id(id)

    def change_team_role(self, user_id, team_id, role_id):
        '''
        Change user role in team
        :param user_id: int
        :param team_id: int
        :param role_id: int
        '''
        response = self._api.post('members.editInfo', {ApiField.ID: user_id,
                                                       ApiField.TEAM_ID: team_id,
                                                       ApiField.ROLE_ID: role_id})

    def get_team_members(self, team_id):
        '''
        :param team_id: int
        :return: list all team users with corresponding roles
        '''
        team_members = self.get_list_all_pages('members.list', {ApiField.TEAM_ID: team_id, ApiField.FILTER: []},
                                               convert_json_info_cb=self._api.user._convert_json_info)
        return team_members

    def get_team_role(self, user_id, team_id):
        '''
        Checks if user with given belongs to team with given id
        :param user_id: int
        :param team_id: int
        '''
        user_teams = self.get_teams(user_id)
        for member in user_teams:
            if member.id == team_id:
                return member
        return None

    def get_member_activity(self, team_id, user_id, progress_cb=None):
        '''
        :param team_id: int
        :param user_id: int
        :param progress_cb: fn
        :return: pandas dataframe (table with activity data of given user)
        '''
        import pandas as pd
        activity = self._api.team.get_activity(team_id, filter_user_id=user_id, progress_cb=progress_cb)
        df = pd.DataFrame(activity)
        return df

    def add_to_team_by_login(self, user_login, team_id, role_id):
        response = self._api.post('members.add', {ApiField.LOGIN: user_login,
                                                  ApiField.TEAM_ID: team_id,
                                                  ApiField.ROLE_ID: role_id})
        return response.json()


# coding: utf-8
import json
import urllib.parse
from functools import partial
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from supervisely.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely.api.video.video_annotation_api import VideoAnnotationAPI
from supervisely.api.video.video_object_api import VideoObjectApi
from supervisely.api.video.video_figure_api import VideoFigureApi
from supervisely.api.video.video_frame_api import VideoFrameAPI
from supervisely.api.video.video_tag_api import VideoTagApi
from supervisely.sly_logger import logger
from supervisely.io.fs import get_file_hash

from supervisely.io.fs import ensure_base_path
from supervisely._utils import batched
from supervisely.video.video import get_video_streams, gen_video_stream_name

from supervisely.task.progress import Progress


class VideoApi(RemoveableBulkModuleApi):
    def __init__(self, api):
        '''
        :param api: Api class object
        '''
        super().__init__(api)
        self.annotation = VideoAnnotationAPI(api)
        self.object = VideoObjectApi(api)
        self.frame = VideoFrameAPI(api)
        self.figure = VideoFigureApi(api)
        self.tag = VideoTagApi(api)

    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.HASH,
                ApiField.TEAM_ID,
                ApiField.WORKSPACE_ID,
                ApiField.PROJECT_ID,
                ApiField.DATASET_ID,
                ApiField.PATH_ORIGINAL,
                ApiField.FRAMES_TO_TIMECODES,
                ApiField.FRAMES_COUNT,
                ApiField.FRAME_WIDTH,
                ApiField.FRAME_HEIGHT,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT,
                ApiField.TAGS,
                ApiField.FILE_META,
                ApiField.CUSTOM_DATA]

    @staticmethod
    def info_tuple_name():
        return 'VideoInfo'

    def url(self, dataset_id, video_id, video_frame=0):
        '''
        :param dataset_id: int
        :param video_id: int
        :param video_frame: int
        :return: url of given video id
        '''
        result = urllib.parse.urljoin(self._api.server_address, f'app/videos/?'
                                                                f'datasetId={dataset_id}&'
                                                                f'videoFrame={video_frame}&'
                                                                f'videoId={video_id}')

        return result

    def _convert_json_info(self, info: dict, skip_missing=True):
        return super(VideoApi, self)._convert_json_info(info, skip_missing=skip_missing)

    def get_list(self, dataset_id, filters=None):
        '''
        :param dataset_id: int
        :param filters: list
        :return: List of the videos from the dataset with given id
        '''
        return self.get_list_all_pages('videos.list',  {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: VideoApi metadata by numeric id
        '''
        return self._get_info_by_id(id, 'videos.info')

    def get_destination_ids(self, id):
        '''
        :param id: int
        :return: id of project and id of dataset from given id
        '''
        dataset_id = self._api.video.get_info_by_id(id).dataset_id
        project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
        return project_id, dataset_id

    def upload_hash(self, dataset_id, name, hash, stream_index=None):
        meta = {}
        if stream_index is not None and type(stream_index) is int:
            meta = {"videoStreamIndex": stream_index}
        return self.upload_hashes(dataset_id, [name], [hash], [meta])[0]

    def upload_hashes(self, dataset_id, names, hashes, metas=None, progress_cb=None):
        return self._upload_bulk_add(lambda item: (ApiField.HASH, item), dataset_id, names, hashes, metas, progress_cb)

    def _upload_bulk_add(self, func_item_to_kv, dataset_id, names, items, metas=None, progress_cb=None):
        if metas is None:
            metas = [{}] * len(items)

        results = []
        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError("Can not match \"names\" and \"items\" lists, len(names) != len(items)")

        for batch in batched(list(zip(names, items, metas))):
            images = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                images.append({'title': name,
                               item_tuple[0]: item_tuple[1],
                               ApiField.META: meta if meta is not None else {}})
            response = self._api.post('videos.bulk.add', {ApiField.DATASET_ID: dataset_id, ApiField.VIDEOS: images})
            if progress_cb is not None:
                progress_cb(len(images))

            results = [self._convert_json_info(item) for item in response.json()]
            name_to_res = {img_info.name: img_info for img_info in results}
            ordered_results = [name_to_res[name] for name in names]

            return ordered_results

    def _download(self, id, is_stream=False):
        '''
        :param id: int
        :param is_stream: bool
        :return: Response object containing video with given id
        '''
        response = self._api.post('videos.download', {ApiField.ID: id}, stream=is_stream)
        return response

    def download_path(self, id, path, progress_cb=None):
        '''
        Download video with given id on the given path
        :param id: int
        :param path: str
        :param progress_cb: SLY Progress Bar callback
        '''
        response = self._download(id, is_stream=True)
        ensure_base_path(path)

        with open(path, 'wb') as fd:
            mb1 = 1024 * 1024
            for chunk in response.iter_content(chunk_size=mb1):
                fd.write(chunk)

                if progress_cb is not None:
                    progress_cb(len(chunk))

    def download_range_by_id(self, id, frame_start, frame_end, is_stream=True):
        '''
        :param id: int
        :param frame_start: int
        :param frame_end: int
        :param is_stream: bool
        :return: Response object containing video with given id between given start and end frames
        '''
        path_original = self.get_info_by_id(id).path_original
        return self.downalod_range_by_path(path_original, frame_start, frame_end, is_stream)

    def downalod_range_by_path(self, path_original, frame_start, frame_end, is_stream=False):
        '''
        :param path_original: str
        :param frame_start: int
        :param frame_end: int
        :param is_stream: bool
        :return: Response object containing video on given path between given start and end frames
        '''
        response = self._api.get(method = 'image-converter/transcode' + path_original,
                                 params={'startFrame': frame_start, 'endFrame': frame_end, "transmux": True},
                                 stream=is_stream,
                                 use_public_api=False)
        return response

    def download_save_range(self, video_id, frame_start, frame_end, save_path):
        '''
        Download video with given id between given start and end frames on the given path
        :param video_id: int
        :param frame_start: int
        :param frame_end: int
        :param save_path:
        :return: str
        '''
        response = self.download_range_by_id(video_id, frame_start, frame_end)
        with open(save_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=128):
                fd.write(chunk)
        return save_path

    def notify_progress(self, track_id, video_id, frame_start, frame_end, current, total):
        '''
        :param track_id: int
        :param video_id: int
        :param frame_start: int
        :param frame_end: int
        :param current: int
        :param total: int
        :return: str
        '''
        response = self._api.post('videos.notify-annotation-tool', {"type": "videos:fetch-figures-in-range",
                                                                    "data": {
                                                                        ApiField.TRACK_ID: track_id,
                                                                        ApiField.VIDEO_ID: video_id,
                                                                        ApiField.FRAME_RANGE: [frame_start, frame_end],
                                                                        ApiField.PROGRESS: {
                                                                            ApiField.CURRENT: current,
                                                                            ApiField.TOTAL: total
                                                                        }
                                                                    }
                                                                    })
        return response.json()[ApiField.STOPPED]

    def notify_tracking_error(self, track_id, error, message):
        '''
        :param track_id: int
        :param error: str
        :param message: str
        :return: #TODO nothing to return
        '''
        response = self._api.post('videos.notify-annotation-tool', {"type": "videos:tracking-error",
                                                                    "data": {
                                                                        ApiField.TRACK_ID: track_id,
                                                                        ApiField.ERROR: {
                                                                            ApiField.MESSAGE: "{}: {}".format(error, message)
                                                                        }
                                                                    }
                                                                    })
    # def upload(self):
    #     #"/videos.bulk.upload"
    #     pass
    #
    # def upload_path(self, dataset_id, name, path, meta=None):
    #     metas = None if meta is None else [meta]
    #     return self.upload_paths(dataset_id, [name], [path], metas=metas)[0]

    #@TODO: copypaste from image_api
    def check_existing_hashes(self, hashes):
        results = []
        if len(hashes) == 0:
            return results
        for hashes_batch in batched(hashes, batch_size=900):
            response = self._api.post('images.internal.hashes.list', hashes_batch)
            results.extend(response.json())
        return results

    def upload_paths(self, dataset_id, names, paths, progress_cb=None, metas=None, infos=None, item_progress=None):
        def path_to_bytes_stream(path):
            return open(path, 'rb')

        update_headers = False
        if infos is not None:
            if len(infos) != len(names):
                raise ValueError("Infos have to be None or provided for all videos")
            update_headers = True

        if update_headers:
            self._api.add_header("x-skip-processing", "true")

        video_info_results = []
        hashes = [get_file_hash(x) for x in paths]

        self._upload_data_bulk(path_to_bytes_stream, zip(paths, hashes), progress_cb=progress_cb, item_progress=item_progress)
        if update_headers:
            self.upsert_infos(hashes, infos)
            self._api.pop_header("x-skip-processing")
        metas = self._api.import_storage.get_meta_by_hashes(hashes)
        metas2 = [meta["meta"] for meta in metas]

        for name, hash, meta in zip(names, hashes, metas2):
            try:
                all_streams = meta["streams"]
                video_streams = get_video_streams(all_streams)
                for stream_info in video_streams:
                    stream_index = stream_info["index"]

                    #TODO: check is community
                    # if instance_type == sly.COMMUNITY:
                    #     if _check_video_requires_processing(file_info, stream_info) is True:
                    #         warn_video_requires_processing(file_name)
                    #         continue

                    item_name = name
                    info = self._api.video.get_info_by_name(dataset_id, item_name)
                    if info is not None:
                        item_name = gen_video_stream_name(name, stream_index)
                    res = self.upload_hash(dataset_id, item_name, hash, stream_index)
                    video_info_results.append(res)
            except Exception as e:
                logger.warning("File skipped {!r}: error occurred during processing {!r}".format(name, str(e)))
        return video_info_results

    def _upload_uniq_videos_single_req(self, func_item_to_byte_stream, hashes_items_to_upload, progress_cb=None):
        content_dict = {}
        for idx, (_, item) in enumerate(hashes_items_to_upload):
            content_dict["{}-file".format(idx)] = (str(idx), func_item_to_byte_stream(item), 'video/*')
        encoder = MultipartEncoder(fields=content_dict)

        if progress_cb is not None:
            def _callback(monitor, progress):
                progress(monitor.bytes_read)
            callback = partial(_callback, progress=progress_cb)
            monitor = MultipartEncoderMonitor(encoder, callback)
            resp = self._api.post('videos.bulk.upload', monitor)
        else:
            resp = self._api.post('videos.bulk.upload', encoder)

        resp_list = json.loads(resp.text)
        remote_hashes = [d['hash'] for d in resp_list if 'hash' in d]
        if len(remote_hashes) != len(hashes_items_to_upload):
            problem_items = [(hsh, item, resp['errors'])
                             for (hsh, item), resp in zip(hashes_items_to_upload, resp_list) if resp.get('errors')]
            logger.warn('Not all images were uploaded within request.', extra={
                'total_cnt': len(hashes_items_to_upload), 'ok_cnt': len(remote_hashes), 'items': problem_items})
        return remote_hashes

    def _upload_data_bulk(self, func_item_to_byte_stream, items_hashes, retry_cnt=3, progress_cb=None, item_progress=None):
        hash_to_items = {i_hash: item for item, i_hash in items_hashes}

        unique_hashes = set(hash_to_items.keys())
        remote_hashes = set(self.check_existing_hashes(list(unique_hashes)))  # existing -- from server
        if progress_cb:
            progress_cb(len(remote_hashes))
        #pending_hashes = unique_hashes #- remote_hashes #@TODO: only fo debug!
        pending_hashes = unique_hashes - remote_hashes

        for retry_idx in range(retry_cnt):
            # single attempt to upload all data which is not uploaded yet
            for hashes in batched(list(pending_hashes)):
                pending_hashes_items = [(h, hash_to_items[h]) for h in hashes]
                hashes_rcv = self._upload_uniq_videos_single_req(func_item_to_byte_stream, pending_hashes_items, item_progress)
                pending_hashes -= set(hashes_rcv)
                if set(hashes_rcv) - set(hashes):
                    logger.warn('Hash inconsistency in images bulk upload.',
                                extra={'sent': hashes, 'received': hashes_rcv})
                if progress_cb:
                    progress_cb(len(hashes_rcv))

            if not pending_hashes:
                return

            logger.warn('Unable to upload videos (data).', extra={
                'retry_idx': retry_idx,
                'items': [(h, hash_to_items[h]) for h in pending_hashes]
            })
            # now retry it for the case if it is a shadow server/connection error

        raise RuntimeError("Unable to upload videos (data). "
                           "Please check if videos are in supported format and if ones aren't corrupted.")

    # @TODO: add case to documentation with detailed explanation
    def upsert_info(self, hash, info):
        return self.upsert_infos([hash], [info])

    def upsert_infos(self, hashes, infos, links=None):
        payload = []
        if links is None:
            links = [None] * len(hashes)
        for h, l, info in zip(hashes, links, infos):
            item = {ApiField.HASH: h, ApiField.META: info}
            if l is not None:
                item[ApiField.LINK] = l
            payload.append(item)

        resp = self._api.post('videos.bulk.upsert_file_meta', payload)
        return resp.json()

    def upload_links(self, dataset_id, names, hashes, links, infos, metas=None):
        self.upsert_infos(hashes, infos, links)
        return self._upload_bulk_add(lambda item: (ApiField.LINK, item), dataset_id, names, links, metas)

    def update_custom_data(self, id: int, data: dict):
        resp = self._api.post('videos.custom-data.set', {ApiField.ID: id, ApiField.CUSTOM_DATA: data})
        return resp.json()



# coding: utf-8

from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectApi(RemoveableBulkModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.DESCRIPTION,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT,
                ApiField.DATASET_ID,
                ApiField.CLASS_ID,
                ApiField.ENTITY_ID,
                ApiField.TAGS,
                ApiField.META,
                ApiField.CREATED_BY_ID,
                ]

    @staticmethod
    def info_tuple_name():
        return 'ObjectInfo'

    def get_info_by_id(self, id):
        return self._get_info_by_id(id, 'objects.info')

    def get_list(self, dataset_id, filters=None):
        return self.get_list_all_pages('annotation-objects.list',  {ApiField.DATASET_ID: dataset_id,
                                                                    ApiField.FILTER: filters or []})

    def _append_bulk(self, tag_api, entity_id, project_id, dataset_id, objects, key_id_map: KeyIdMap = None, is_pointcloud=False):
        if len(objects) == 0:
            return []

        objcls_name_id_map = self._api.object_class.get_name_to_id_map(project_id)

        items = []
        for obj in objects:
            new_obj = {ApiField.CLASS_ID: objcls_name_id_map[obj.obj_class.name]}

            if not is_pointcloud:
            #if entity_id is not None:
                new_obj[ApiField.ENTITY_ID] = entity_id
            items.append(new_obj)

        response = self._api.post('annotation-objects.bulk.add', {ApiField.DATASET_ID: dataset_id,
                                                                  ApiField.ANNOTATION_OBJECTS: items})
        ids = [obj[ApiField.ID] for obj in response.json()]
        KeyIdMap.add_objects_to(key_id_map, [obj.key() for obj in objects], ids)

        # add tags to objects
        tag_api.append_to_objects(entity_id, project_id, objects, key_id_map)

        return ids



# coding: utf-8

from copy import deepcopy

from supervisely.io.json import JsonSerializable
from supervisely.imaging import image as sly_image
from supervisely.geometry import validation
from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS
from supervisely._utils import unwrap_if_numpy


class PointLocation(JsonSerializable):
    def __init__(self, row, col):
        """
        Create simple point in (row, col) position. Float-type coordinates will be deprecated soon.
        """
        self._row = round(unwrap_if_numpy(row))
        self._col = round(unwrap_if_numpy(col))

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    def to_json(self):
        '''
        The function to_json convert PointLocation class object to json format
        :return: PointLocation in json format
        '''
        packed_obj = {
            POINTS: {
                EXTERIOR: [[self.col, self.row]],
                INTERIOR: []
            }
        }
        return packed_obj

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert PointLocation from json format to PointLocation class object.
        :param data: input PointLocation in json format
        :return: PointLocation class object
        '''
        validation.validate_geometry_points_fields(data)
        exterior = data[POINTS][EXTERIOR]
        if len(exterior) != 1:
            raise ValueError('"exterior" field must contain exactly one point to create "PointLocation" object.')
        return cls(row=exterior[0][1], col=exterior[0][0])

    def scale(self, factor):
        '''
        The function scale calculates new parameters of point after scaling image
        :param factor: float scale parameter
        :return: PointLocation class object
        '''
        return self.scale_frow_fcol(factor, factor)

    def scale_frow_fcol(self, frow, fcol):
        '''
        The function scale_frow_fcol calculates new parameters of point after scaling with given parameters in horizontal and vertical
        :param frow: float scale parameter
        :param fcol: float scale parameter
        :return: PointLocation class object
        '''
        return PointLocation(row=round(self.row * frow), col=round(self.col * fcol))

    def translate(self, drow, dcol):
        '''
        The function translate calculates new parameters of point after shifts it by a certain number of pixels
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: PointLocation class object
        '''
        return PointLocation(row=(self.row + drow), col=(self.col + dcol))

    def rotate(self, rotator):
        '''
        The function rotate calculates new parameters of point after rotating
        :param rotator: ImageRotator class object
        :return: PointLocation class object
        '''
        return rotator.transform_point(self)

    def resize(self, in_size, out_size):
        '''
        The function resize calculates new parameters of point after resizing image
        :param in_size: input image size
        :param out_size: output image size
        :return: PointLocation class object
        '''
        new_size = sly_image.restore_proportional_size(in_size=in_size, out_size=out_size)
        frow = new_size[0] / in_size[0]
        fcol = new_size[1] / in_size[1]
        return self.scale_frow_fcol(frow=frow, fcol=fcol)

    def fliplr(self, img_size):
        '''
        The function fliplr calculates new parameters of point after fliping image in horizontal
        :param img_size: image size
        :return: PointLocation class object
        '''
        return PointLocation(row=self.row, col=(img_size[1] - self.col))

    def flipud(self, img_size):
        '''
        The function flipud calculates new parameters of point after fliping image in vertical
        :param img_size: image size
        :return: PointLocation class object
        '''
        return PointLocation(row=(img_size[0] - self.row), col=self.col)

    def clone(self):
        return deepcopy(self)


def _flip_row_col_order(coords):
    if not all(len(x) == 2 for x in coords):
        raise ValueError('Flipping row and column order values is only possible within tuples of 2 elements.')
    return [[y, x] for x, y in coords]


def _maybe_flip_row_col_order(coords, flip=False):
    return _flip_row_col_order(coords) if flip else coords


def points_to_row_col_list(points, flip_row_col_order=False):
    '''
    The function points_to_row_col_list convert list of points(PointLocation objects) to list of coords, flip row col coords if flip_row_col_order==True
    :param points: list of points(PointLocation class objects)
    :param flip_row_col_order: bool param for flipping coords
    :return: list of coords
    '''
    return _maybe_flip_row_col_order(coords=[[p.row, p.col] for p in points], flip=flip_row_col_order)


def row_col_list_to_points(data, flip_row_col_order=False, do_round=False):
    '''
    The function row_col_list_to_points convertlist of coords to list of points(PointLocation objects), flip row col coords if flip_row_col_order==True, round PointLocation params if do_round==True
    :param data: list of coords
    :param flip_row_col_order: bool param for flipping coords
    :param do_round: bool param to round coords
    :return: list of points(PointLocation class objects)
    '''
    def _maybe_round(v):
        return v if not do_round else round(v)

    return [PointLocation(row=_maybe_round(r), col=_maybe_round(c)) for r, c in
            _maybe_flip_row_col_order(data, flip=flip_row_col_order)]




import enum
import json
import jsonpatch
import asyncio
from fastapi import FastAPI
from fastapi import Request
from supervisely.fastapi_helpers.singleton import Singleton
from supervisely.fastapi_helpers.websocket import WebsocketManager


class Field(str, enum.Enum):
    STATE = 'state'
    DATA = 'data'
    CONTEXT = 'context'


class _PatchableJson(dict):
    def __init__(self, field: Field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ws = WebsocketManager()
        self._last = dict(self)
        self._lock = asyncio.Lock()
        self._field = field.value

    def get_changes(self, patch=None):
        if patch is None:
            patch = self._get_patch()
        return {self._field: json.loads(patch.to_string())}

    def _get_patch(self):
        patch = jsonpatch.JsonPatch.from_diff(self._last, self)
        return patch

    async def _apply_patch(self, patch):
        async with self._lock:
            patch.apply(self._last, in_place=True)

    async def synchronize_changes(self):
        patch = self._get_patch()
        await self._apply_patch(patch)
        await self._ws.broadcast(self.get_changes(patch))

    @classmethod
    async def _from_request(cls, field, request: Request):
        content = await request.json()
        d = content.get(field, {})
        return cls(d)
    
    @classmethod
    async def from_request(cls, request: Request):
        raise NotImplementedError()


class LastStateJson(_PatchableJson, metaclass=Singleton):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.STATE, *args, **kwargs)
    
    @classmethod
    async def from_request(cls, request: Request):
        content = await request.json()
        last_state = cls()
        d = content.get(last_state._field)
        if d is not None:
            async with last_state._lock:
                last_state.clear()
                last_state.update(d)
        return last_state
    
    async def replace(self, d: dict): 
        # update method already exists in dict
        if d is not None:  
            async with self._lock:
                self.clear()
                self.update(d)


class ContextJson(_PatchableJson):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.CONTEXT, *args, **kwargs)
    
    @classmethod
    async def from_request(cls, request: Request):
        return await cls._from_request(Field.CONTEXT, request)


class StateJson(_PatchableJson):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.STATE, *args, **kwargs)
    
    async def _apply_patch(self, patch):
        await super()._apply_patch(patch)
        await LastStateJson()._apply_patch(patch)

    @classmethod
    async def from_request(cls, request: Request):
        state_json = await cls._from_request(Field.STATE, request)
        await LastStateJson().replace(state_json)
        return state_json


class DataJson(_PatchableJson, metaclass=Singleton):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.DATA, *args, **kwargs)
    
    @classmethod
    async def from_request(cls, request: Request):
        raise RuntimeError(f"""Request from Supervisely App never contains \"{cls._field}\" field. Every request from app contains by default current state and context""")





# coding: utf-8
import base64
import zlib
import io
import numpy as np

from supervisely.geometry.bitmap_base import BitmapBase, resize_origin_and_bitmap
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.constants import MULTICHANNEL_BITMAP


class MultichannelBitmap(BitmapBase):
    @staticmethod
    def geometry_name():
        return 'multichannelBitmap'

    def __init__(self, data, origin: PointLocation = None,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        The constructor for MultichannelBitmap class.
        :param data: bool numpy array
        :param origin: points (x and y coordinates) of the top left corner of a bitmap, i.e. the position of the
        bitmap within the image
        '''
        super().__init__(data, origin, expected_data_dims=3,
                         sly_id=sly_id, class_id=class_id,
                         labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    @classmethod
    def _impl_json_class_name(cls):
        return MULTICHANNEL_BITMAP

    def rotate(self, rotator):
        '''
        The function rotate render the bitmap within the full image canvas and rotate the whole canvas
        with a given rotator (ImageRotator class object contain size of image and angle to rotate)
        :param rotator: ImageRotator class object
        :return: MultichannelBitmap class object
        '''
        full_img_data = np.zeros(rotator.src_imsize + self.data.shape[2:], dtype=self.data.dtype)
        full_img_data[
            self.origin.row:(self.origin.row + self.data.shape[0]),
            self.origin.col:(self.origin.col + self.data.shape[1]), ...] = self.data[:, :, ...]
        rotated_full_data = rotator.rotate_img(full_img_data, use_inter_nearest=True)
        # Rotate the bounding box to find out the bounding box of the rotated bitmap within the full image.
        rotated_bbox = self.to_bbox().rotate(rotator)
        rotated_origin = PointLocation(row=rotated_bbox.top, col=rotated_bbox.left)
        return MultichannelBitmap(data=rotated_bbox.get_cropped_numpy_slice(rotated_full_data), origin=rotated_origin)

    def crop(self, rect):
        '''
        Crop the current MultichannelBitmap object with a given rectangle
        :param rect: Rectangle class object
        :return: MultichannelBitmap class object
        '''
        maybe_cropped_area = self.to_bbox().crop(rect)
        if len(maybe_cropped_area) == 0:
            return []
        else:
            [cropped_area] = maybe_cropped_area
            cropped_origin = PointLocation(row=cropped_area.top, col=cropped_area.left)
            cropped_area_in_data = cropped_area.translate(drow=-self._origin.row, dcol=-self.origin.col)
            return [MultichannelBitmap(data=cropped_area_in_data.get_cropped_numpy_slice(self._data),
                                       origin=cropped_origin,)]

    def resize(self, in_size, out_size):
        '''
        Resize the current MultichannelBitmap to match a certain size
        :param in_size: input image size
        :param out_size: output image size
        :return: MultichannelBitmap class object
        '''
        scaled_origin, scaled_data = resize_origin_and_bitmap(self._origin, self._data, in_size, out_size)
        return MultichannelBitmap(data=scaled_data, origin=scaled_origin)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        self.to_bbox().get_cropped_numpy_slice(bitmap)[...] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        bbox = self.to_bbox()
        # Not forwarding the config here directly since a Rectangle cannot know
        # about our config format.
        bbox.draw_contour(bitmap, color, thickness)

    @property
    def area(self):
        '''
        :return: area of current MultichannelBitmap object
        '''
        return self.data.shape[0] * self.data.shape[1]

    @staticmethod
    def base64_2_data(s: str) -> np.ndarray:
        '''
        The function base64_2_data convert base64 encoded string to numpy
        :param s: string
        :return: numpy array
        '''
        saved_bytes = io.BytesIO(zlib.decompress(base64.b64decode(s)))
        return np.load(saved_bytes)

    @staticmethod
    def data_2_base64(data: np.ndarray) -> str:
        '''
        he function data_2_base64 convert numpy array to base64 encoded string
        :param data: numpy array
        :return: string
        '''
        bytes_io = io.BytesIO()
        np.save(bytes_io, data, allow_pickle=False)
        return base64.b64encode(zlib.compress(bytes_io.getvalue())).decode('utf-8')

    def validate(self, name, settings):
        pass  # No need name validation - inner geometry type.



# coding: utf-8
from __future__ import annotations
from prettytable import PrettyTable
from supervisely._utils import take_with_default
from typing import List, Iterable
from collections import defaultdict


class DuplicateKeyError(KeyError):
    r"""Raised when trying to add already existing key to
    :class:`KeyIndexedCollection <supervisely.collection.key_indexed_collection.KeyIndexedCollection>`"""
    pass


class KeyObject:
    r"""Base class fo objects that should implement ``key`` method. Child classes then can be stored
    KeyIndexedCollection.
    """
    def key(self):
        r"""
        Returns:
            can be any immutable type; strings and numbers
        """
        raise NotImplementedError()


class KeyIndexedCollection:
    r"""Base class for ObjClassCollection, TagMetaCollection and TagCollection instances
    It is an analogue of python's standard Dict. It allows to store objects inherited from
    :class:`KeyObject <supervisely.collection.key_indexed_collection.KeyObject>`.
    But it raises :class:`DuplicateKeyError <supervisely.collection.key_indexed_collection.DuplicateKeyError>`
    exception when trying to add object with already existing key.
    """
    
    item_type = KeyObject
    """The type of items that can be storred in collection. Defaul value is 
    :class:`KeyObject <supervisely.collection.key_indexed_collection.KeyObject>`. 
    Field has to be overridden in child class. Before adding object to collection its type is compared with 
    ``item_type`` and ``TypeError`` exception is raised if it differs. Collection is immutable.
    """

    def __init__(self, items=None):
        '''
        :param items: dictionary containing collection in format name -> item
        '''
        self._collection = {}
        self._add_items_impl(self._collection, take_with_default(items, []))

    def _add_impl(self, dst_collection, item):
        '''
        Add given item to given collection. Raise error if type of item not KeyObject or item with an item with that name is already in given collection
        :param dst_collection: dictionary
        :param item: ObjClass, TagMeta or Tag class object
        :return: dictionary
        '''
        if not isinstance(item, KeyIndexedCollection.item_type):
            raise TypeError(
                'Item type ({!r}) != {!r}'.format(type(item).__name__, KeyIndexedCollection.item_type.__name__))
        if item.key() in dst_collection:
            raise DuplicateKeyError('Key {!r} already exists'.format(item.key()))
        dst_collection[item.key()] = item

    def _add_items_impl(self, dst_collection, items):
        '''
        Add items from input list to given collection. Raise error if type of item not KeyObject or item with an item with that name is already in given collection
        :param dst_collection: dictionary
        :param items: list of ObjClass, TagMeta or Tag class objects
        '''
        for item in items:
            self._add_impl(dst_collection, item)

    def add(self, item):
        '''
        Add given item to collection
        :param item: ObjClass, TagMeta or Tag class object
        :return: KeyIndexedCollection class object
        '''
        return self.clone(items=[*self.items(), item])

    def add_items(self, items):
        '''
        Add items from given list to collection
        :param items: list of ObjClass, TagMeta or Tag class objects
        :return: KeyIndexedCollection class object
        '''
        """Generators have a ``Yields`` section instead of a ``Returns`` section.

        Args:
            n (int): The upper limit of the range to generate, from 0 to `n` - 1.

        Yields:
            int: The next number in the range of 0 to `n` - 1.

        Examples:
            Examples should be written in doctest format, and should illustrate how
            to use the function.

            >>> print([i for i in example_generator(4)])
            [0, 1, 2, 3]

        """
        return self.clone(items=[*self.items(), *items])

    def get(self, key, default=None):
        '''
        The function get return item from collection with given key(name)
        :param key: str
        :return: ObjClass, TagMeta or Tag class object
        '''
        return self._collection.get(key, default)

    def __next__(self):
        for value in self._collection.values():
            yield value

    def __iter__(self):
        return next(self)

    def __contains__(self, item):
        return (isinstance(item, KeyIndexedCollection.item_type)
                and item == self._collection.get(item.key()))

    def __len__(self):
        return len(self._collection)

    def items(self):
        '''
        :return: list of all items in collection
        '''
        return list(self._collection.values())

    def clone(self, items=None):
        '''
        :param items: list of ObjClass, TagMeta or Tag class objects
        :return: KeyIndexedCollection class object with given items in collection, if given list of items is None - return copy of KeyIndexedCollection
        '''
        return type(self)(items=(items if items is not None else self.items()))

    def keys(self):
        '''
        :return: list of all keys(item names) in collection
        '''
        return list(self._collection.keys())

    def has_key(self, key):
        '''
        Check if given key(item name exist in collection)
        :param key: str
        :return: bool
        '''
        """Exceptions are documented in the same way as classes.

        The __init__ method may be documented in either the class level
        docstring, or as a docstring on the __init__ method itself.

        Either form is acceptable, but the two should not be mixed. Choose one
        convention to document the __init__ method and be consistent with it.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            msg (str): Human readable string describing the exception.
            code (:obj:`int`, optional): Error code.

        Attributes:
            msg (str): Human readable string describing the exception.
            code (int): Exception error code.

        """
        return key in self._collection

    def intersection(self, other):
        '''
        The function intersection find intersection of given list of instances with collection items. Raise error if find items with same keys(item names)
        :param other: list of ObjClass, TagMeta or Tag class objects
        :return: KeyIndexedCollection class object
        '''
        common_items = []
        for other_item in other:
            our_item = self.get(other_item.key())
            if our_item is not None:
                if our_item != other_item:
                    raise ValueError("Different values for the same key {!r}".format(other_item.key()))
                else:
                    common_items.append(our_item)
        return self.clone(common_items)

    def difference(self, other):
        '''
        The function difference find difference between collection and given list of instances
        :param other: list of ObjClass, TagMeta or Tag class objects
        :return: KeyIndexedCollection class object
        '''
        items = [item for item in self.items() if item not in other]
        return self.clone(items)

    def merge(self, other):
        '''
        The function merge merge collection and given list of instances. Raise error if item name from given list is in collection but items in both are different
        :param other: list of ObjClass, TagMeta or Tag class objects
        :return: KeyIndexedCollection class object
        '''
        new_items = []
        for other_item in other.items():
            our_item = self.get(other_item.key())
            if our_item is None:
                new_items.append(other_item)
            elif our_item != other_item:
                raise ValueError('Error during merge for key {!r}: values are different'.format(other_item.key()))
        return self.clone(new_items + self.items())

    def __str__(self):
        res_table = PrettyTable()
        res_table.field_names = self.item_type.get_header_ptable()
        for item in self:
            res_table.add_row(item.get_row_ptable())
        return res_table.get_string()

    def to_json(self) -> List[dict]:
        """
        Converts collection to json serializable list.
        Returns:
            json serializable dictionary
        """
        return [item.to_json() for item in self]

    def __eq__(self, other: KeyIndexedCollection):
        if len(self) != len(other):
            return False
        for cur_item in self:
            other_item = other.get(cur_item.key())
            if other_item is None or cur_item != other_item:
                return False
        return True

    def __ne__(self, other: KeyIndexedCollection):
        return not self == other


class MultiKeyIndexedCollection(KeyIndexedCollection):
    def __init__(self, items=None):
        self._collection = defaultdict(list)
        self._add_items_impl(self._collection, take_with_default(items, []))

    def _add_impl(self, dst_collection, item):
        if not isinstance(item, MultiKeyIndexedCollection.item_type):
            raise TypeError(
                'Item type ({!r}) != {!r}'.format(type(item).__name__, MultiKeyIndexedCollection.item_type.__name__))
        dst_collection[item.key()].append(item)

    def get(self, key, default=None):
        result = self._collection.get(key, default)
        if not result:
            return None
        return result[0]

    def get_all(self, key, default=[]):
        return self._collection.get(key, default)

    def __next__(self):
        for tag_list in self._collection.values():
            for tag in tag_list:
                yield tag

    def __contains__(self, item):
        return (isinstance(item, MultiKeyIndexedCollection.item_type)
                and item in self.get_all(item.key()))

    def __len__(self):
        return sum([len(tag_list) for tag_list in self._collection.values()])

    def items(self):
        res = []
        for tag_list in self._collection.values():
            res.extend(tag_list)
        return res

    def intersection(self, other):
        common_items = []
        for other_item in other:
            key_list = self.get_all(other_item.key())
            for our_item in key_list:
                if our_item == other_item:
                    common_items.append(our_item)
        return self.clone(common_items)

    def merge(self, other):
        new_items = [*self.items(), *other.items()]
        return self.clone(items=new_items)

    def merge_without_duplicates(self, other):
        return super().merge(other)



from supervisely.fastapi_helpers.websocket import WebsocketManager
from supervisely.fastapi_helpers.templating import Jinja2Templates
from supervisely.fastapi_helpers.app_content import StateJson, DataJson, LastStateJson, ContextJson
from supervisely.fastapi_helpers.subapp import get_subapp, shutdown


import os
import signal
import psutil

# https://fastapi.tiangolo.com/advanced/sub-applications/
from fastapi import FastAPI, Request, Response, WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse

from supervisely.fastapi_helpers import DataJson, LastStateJson
from supervisely.fastapi_helpers import WebsocketManager


def get_subapp() -> FastAPI:
    app = FastAPI()
    WebsocketManager().set_app(app)

    @app.post("/data")
    async def send_data(request: Request):
        data = DataJson()
        response = JSONResponse(content=dict(data))
        return response

    @app.post("/state")
    async def send_state(request: Request):
        state = LastStateJson()
        response = JSONResponse(content=dict(state))
        return response

    @app.post("/shutdown")
    async def shutdown_endpoint(request: Request):
        shutdown()

    @app.websocket("/ws")
    async def websocket_endpoint(websocket: WebSocket):
        await WebsocketManager().connect(websocket)
        try:
            while True:
                data = await websocket.receive_text()
        except WebSocketDisconnect:
            WebsocketManager().disconnect(websocket)

    return app


def shutdown():
    current_process = psutil.Process(os.getpid())
    current_process.send_signal(signal.SIGINT) # emit ctrl + c